#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#include "../../ogg.h"
#include "../../unzip.h"
//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#include <condefs.h>
#pragma hdrstop
USERC("compres.rc");
USEUNIT("..\..\ogg.cpp");
USEUNIT("..\..\unzip.cpp");
#endif
//---------------------------------------------------------------------------
HINSTANCE hInstance;


void *LoadOggWav(HINSTANCE h,LPCTSTR name,UINT fuload);
void DeleteWav(void *wav);
bool GetWav(void *wav, WAVEFORMATEX *wfx, unsigned long *size,unsigned long *numsamples);
// LoadOggWav - loads a ogg either from a file (fuload=LR_LOADFROMFILE)
// or from a resource (type=RCDATA,id=name) (fuload=0)
// The returned value is a WAV file laid out in memory, from where
// you can PlaySound(...,SND_MEMORY) it, or save it to a wav file on disk.
// GetWav - returns information about it. size is the entire size of the
// wav file, should you wish to save it to disk.
HBITMAP LoadZipImage(HINSTANCE h, LPCSTR zipname, LPCSTR imagename,UINT fuload);
// LoadZipImage - loads a bmp from a zip file (fuload=LR_LOADFROMFILE)
// or from a zip resource (type=RCDATA,id=zipname) (fuload=0)
// The returned bitmap is either a DIBSection (fuload|=LR_CREATEDIBSECTION)
// or a device-dependent bitmap (otherwise).


void *wav;
HANDLE hbm[16];
int bframe=0,bx=10,by=10,dirx=2,diry=2,bwidth,bheight;
unsigned int nextplay=0,nextpos=0,nextframe=0;
unsigned long wavms;


LRESULT CALLBACK PlainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{ switch (msg)
  { case WM_CREATE:
    { // load our bitmap images from the resource zipfile
      for (int i=0; i<16; i++)
      { char c[30]; wsprintf(c,"frame%i.bmp",i+1);
        hbm[i]=LoadZipImage(hInstance,"zips",c,0);
      }
      BITMAP bmp; GetObject(hbm[0],sizeof(bmp),&bmp);
      bwidth=bmp.bmWidth; bheight=bmp.bmHeight;
      // load the music from the resource oggfile
      wav = LoadOggWav(hInstance,"music",0);
      WAVEFORMATEX wfx; unsigned long nsamples; GetWav(wav,&wfx,NULL,&nsamples);
      wavms = nsamples*1000/wfx.nSamplesPerSec;
      // ready to go!
      SetTimer(hwnd,1,10,NULL);
    } break;
    case WM_TIMER:
    { unsigned int time=GetTickCount();
      if (time>nextplay)
      { PlaySound((const char*)wav,NULL,SND_MEMORY|SND_ASYNC);
        nextplay = time+wavms;
      }
      if (time>nextpos)
      { RECT rc; GetClientRect(hwnd,&rc);
        bx+=dirx; by+=diry;
        if (dirx>0 && bx+bwidth>rc.right) dirx=-dirx;
        if (dirx<0 && bx<0) dirx=-dirx;
        if (diry>0 && by+bheight>rc.bottom) diry=-diry;
        if (diry<0 && by<0) diry=-diry;
        InvalidateRect(hwnd,NULL,TRUE);
        nextpos = time+50;
      }
      if (time>nextframe)
      { bframe = (bframe+1)%16;
        InvalidateRect(hwnd,NULL,TRUE);
        nextframe = time+100;
      }
    } break;
    case WM_PAINT:
    { PAINTSTRUCT ps; BeginPaint(hwnd,&ps);
      HDC tdc = CreateCompatibleDC(ps.hdc);
      HGDIOBJ hold = SelectObject(tdc,hbm[bframe]);
      BitBlt(ps.hdc,bx,by,bwidth,bheight,tdc,0,0,SRCCOPY);
      SelectObject(tdc,hold);
      DeleteDC(tdc);
      EndPaint(hwnd,&ps);
    } break;
    case WM_CLOSE:
    { DestroyWindow(hwnd);
    } break;
    case WM_DESTROY:
    { KillTimer(hwnd,1);
      DeleteWav(wav); wav=0;
      for (int i=0; i<16; i++) {DeleteObject(hbm[i]); hbm[i]=0;}
      PostQuitMessage(0);
    } break;
  }
  return DefWindowProc(hwnd, msg, wParam, lParam);
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ hInstance=h;
  //
  WNDCLASSEX wcex; ZeroMemory(&wcex,sizeof(wcex)); wcex.cbSize = sizeof(WNDCLASSEX);
  BOOL res=GetClassInfoEx(hInstance,"PlainClass",&wcex);
  if (!res)
  { wcex.style = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc = (WNDPROC)PlainWndProc;
    wcex.cbClsExtra = 0;
    wcex.cbWndExtra = 0;
    wcex.hInstance = hInstance;
    wcex.hIcon = NULL;
    wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
    wcex.lpszMenuName	= NULL;
    wcex.lpszClassName = "PlainClass";
    wcex.hIconSm = NULL;
    ATOM res=RegisterClassEx(&wcex);
    if (res==0) {MessageBox(NULL,"Failed to register class","Error",MB_OK); return 0;}
  }
  //
  HWND hwnd = CreateWindowEx(0,"PlainClass", "Compressed Resources", WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT, 400, 400, NULL, NULL, hInstance, NULL);
  if (hwnd==NULL) {MessageBox(NULL,"Failed to create window","Error",MB_OK); return 0;}
  ShowWindow(hwnd,SW_SHOW);
  //
  MSG msg;
  while (GetMessage(&msg, NULL, 0, 0))
  { TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
  return (int)msg.wParam;
}





// ----------------------------------------------------------------------

// The WAV file format.
// The format is a single RIFF WAVE chunk, groupid="RIFF", rifftype="WAVE"
// This contains two chunks, first the format chunk ckID="fmt "
// and next the sound data chunk ckID="data"
// Note: all RIFF blocks contain 4byte identifier, then 4byte 'sizeofrest', then the rest of the chunk.
// The following structures represent the structure of a WAV file.
// NB. the use of pshpack1.h, to ensure that the structures don't get padded by the compiler.
#include <pshpack1.h>
typedef struct
{ char id[4];         //="fmt "
  unsigned long size; //=16
  short wFormatTag;   //=WAVE_FORMAT_PCM=1
  unsigned short wChannels;       //=1 or 2 for mono or stereo
  unsigned long dwSamplesPerSec;  //=11025 or 22050 or 44100
  unsigned long dwAvgBytesPerSec; //=wBlockAlign * dwSamplesPerSec
  unsigned short wBlockAlign;     //=wChannels * (wBitsPerSample==8?1:2)
  unsigned short wBitsPerSample;  //=8 or 16, for bits per sample
} FmtChunk;

typedef struct
{ char id[4];            //="data"
  unsigned long size;    //=datsize, size of the following array
  char data[1];          //=the raw data goes here
} DataChunk;

typedef struct
{ char id[4];         //="RIFF"
  unsigned long size; //=datsize+8+16+4
  char type[4];       //="WAVE"
  FmtChunk fmt;
  DataChunk dat;
} WavChunk;
#include <poppack.h>



// MFILE -- acts like a FILE, but for memory.
typedef struct {unsigned int pos,size; const char *buf;} MFILE;
MFILE *mopen(void *buf,unsigned int size)
{ MFILE *m = new MFILE;
  m->size=size; m->buf=(const char*)buf; m->pos=0;
  return m;
}
void mclose(MFILE *m)
{ delete m;
}
size_t mread_ogg(void *ptr,size_t size,size_t nmemb,void *src)
{ MFILE *m = (MFILE*)src; if (m==0) return 0;
  if (m->pos>m->size) return 0;
  unsigned int wanted = (unsigned int)size*nmemb;
  unsigned int actual = wanted; if (m->pos+actual>m->size) actual=m->size-m->pos;
  memcpy(ptr,m->buf+m->pos,actual);
  m->pos += actual;
  return actual;
}
int mseek_ogg(void *src,__int64 offset,int whence)
{ MFILE *m = (MFILE*)src; if (m==0) return -1;
  if (whence==SEEK_CUR) {m->pos+=(unsigned int)offset; return 0;}
  else if (whence==SEEK_END) {m->pos=m->size+(unsigned int)offset; return 0;}
  else if (whence==SEEK_SET) {m->pos=(unsigned int)offset; return 0;}
  else return -1;
}
int mclose_ogg(void *src)
{ MFILE *m = (MFILE*)src; if (m==0) return -1;
  mclose(m); return 0;
}
long mtell_ogg(void *src)
{ MFILE *m=(MFILE*)src; if (m==0) return -1;
  return m->pos;
}
const ov_callbacks mfile_callbacks = {mread_ogg,mseek_ogg,mclose_ogg,mtell_ogg};




void *LoadOggWav(HINSTANCE hinst,LPCTSTR name,UINT fuload)
{ if (fuload!=LR_LOADFROMFILE && fuload!=0) return 0;

  OggVorbis_File vf;
  if (fuload==LR_LOADFROMFILE)
  { FILE *fin = fopen(name,"rb");
    if (fin==NULL) return 0;
    int res=ov_open(fin,&vf,NULL,0);
    if (res<0) {fclose(fin); return 0;}
  }
  else
  { HRSRC hrsrc = FindResource(hinst,name,RT_RCDATA); if (hrsrc==NULL) return 0;
    HGLOBAL hglob = LoadResource(hinst,hrsrc); if (hglob==0) return 0;
    DWORD size=SizeofResource(hinst,hrsrc);
    void *oggbuf = LockResource(hglob);
    // no need to free it!
    MFILE *min = mopen(oggbuf,size);
    if (min==NULL) return 0;
    int res=ov_open_callbacks(min,&vf,NULL,0,mfile_callbacks);
    if (res<0) {mclose(min); return 0;}
  }

  // note that, if ov_open succeeds, then vf has ownership of the file (be it
  // disk or resource) and will be responsible for closing it.
  vorbis_info *vi=ov_info(&vf,-1);
  if (!ov_seekable(&vf)) {ov_clear(&vf); return 0;}
  //
  // Now we will create a WAV format laid out in memory.
  // It's possible (and easy!) to play this directly, using
  // the call PlaySound(). But instead, we're just going to
  // write it to disk.
  //
  // Problem: although ogg claims to give us ov_pcm_total, it shortchanges
  // us a bit. Hence we leave some extra safety margin here, and reset
  // the header at the end
  const unsigned int extra=10000;
  unsigned int numsamples = extra+(unsigned int)ov_pcm_total(&vf,0);
  unsigned int numbytes = numsamples*vi->channels*2; // 16bit output wav
  unsigned int wavsize = sizeof(WavChunk)+numbytes;
  char *wav = new char[wavsize]; WavChunk *wc = (WavChunk*)wav;
  // this following is just the layout of a WAV file.
  strncpy(wc->id,"RIFF",4);
  wc->size = 4 + sizeof(FmtChunk)+sizeof(DataChunk)+numbytes;
  strncpy(wc->type,"WAVE",4);
  strncpy(wc->fmt.id,"fmt ",4);
  wc->fmt.size=16;
  wc->fmt.wFormatTag=1;
  wc->fmt.wChannels = (unsigned short)vi->channels;
  wc->fmt.dwSamplesPerSec = vi->rate;
  wc->fmt.wBitsPerSample = 16;
  wc->fmt.wBlockAlign = (unsigned short)(wc->fmt.wChannels * wc->fmt.wBitsPerSample/8);
  wc->fmt.dwAvgBytesPerSec = wc->fmt.wBlockAlign * wc->fmt.dwSamplesPerSec;
  strncpy(wc->dat.id,"data",4);
  wc->dat.size = numbytes;
  char *buf = wc->dat.data; unsigned long bufpos=0;

  // now do the actual translating of raw data
  int current_section=0; for (;;)
  { long ret = ov_read(&vf,buf+bufpos,numbytes-bufpos,0,2,1,&current_section);
    if (ret<=0) break;
    else bufpos+=ret;
  }

  // now recalculate the actual size
  numbytes = bufpos;
  //numsamples = numbytes/2/vi->channels;
  wc->dat.size = numbytes;

  ov_clear(&vf);

  // Now: we have allocated wav[wavsize]
  // of which the first sizeof(WavChunk)+numbytes are the actual WAV file
  // and the rest are junk. And we have numsamples, giving a playtime
  // of milliseconds 1000*numsamples/(unsigned int)wc->fmt.dwSamplesPerSec
  return wav;
}

void DeleteWav(void *buf)
{ char *b = (char*)buf;
  delete[] b;
}

bool GetWav(void *wav, WAVEFORMATEX *wfx, unsigned long *size,unsigned long *nsamples)
{ if (wav==0) return false;
  WavChunk *wc = (WavChunk*)wav;
  if (wfx!=NULL)
  { wfx->wFormatTag=wc->fmt.wFormatTag;
    wfx->nChannels=wc->fmt.wChannels;
    wfx->nSamplesPerSec=wc->fmt.dwSamplesPerSec;
    wfx->nAvgBytesPerSec=wc->fmt.dwAvgBytesPerSec;
    wfx->nBlockAlign=wc->fmt.wBlockAlign;
    wfx->wBitsPerSample=wc->fmt.wBitsPerSample;
    wfx->cbSize=0;
  }
  unsigned long numbytes = wc->dat.size;
  if (size!=0)
  { *size=numbytes+sizeof(WavChunk);
  }
  if (nsamples!=0)
  { *nsamples = numbytes/2/wc->fmt.wChannels;
  }
  return true;
}



// ----------------------------------------------------------------------

// LoadZipImage - loads a bmp from a zip file (fuload=LR_LOADFROMFILE)
// or from a zip resource (type=RCDATA,id=zipname) (fuload=0)
// The returned bitmap is either a DIBSection (fuload|=LR_CREATEDIBSECTION)
// or a device-dependent bitmap (otherwise).
HBITMAP LoadZipImage(HINSTANCE hinst, LPCSTR zipname, LPCSTR imagename,UINT fuload)
{ HZIP hz;
  if ((fuload&LR_LOADFROMFILE)==LR_LOADFROMFILE)
  { HANDLE hf = CreateFile(zipname,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
    hz = OpenZip(hf,0,ZIP_HANDLE);
    CloseHandle(hf);
  }
  else
  { HRSRC hrsrc = FindResource(hinst,zipname,RT_RCDATA); if (hrsrc==NULL) return 0;
    HGLOBAL hglob = LoadResource(hinst,hrsrc); if (hglob==0) return 0;
    DWORD size=SizeofResource(hinst,hrsrc);
    void *zipbuf = LockResource(hglob);
    hz = OpenZip(zipbuf,size,ZIP_MEMORY);
  }
  int index; ZIPENTRY ze; FindZipItem(hz,imagename,true,&index,&ze);
  if (index==-1) return 0; // not found
  char *buf = new char[ze.unc_size];
  UnzipItem(hz,index, buf,ze.unc_size,ZIP_MEMORY);
  CloseZip(hz);
  // now turn it into a bitmap!
  BITMAPFILEHEADER *bfh=(BITMAPFILEHEADER*)buf;
  BITMAPINFOHEADER *bih=(BITMAPINFOHEADER*)(buf+sizeof(BITMAPFILEHEADER));
  //RGBQUAD *cols = (RGBQUAD*)(buf+sizeof(BITMAPFILEHEADER)+bih->biSize);
  char *sbits = (char*)(buf+bfh->bfOffBits);
  HBITMAP hbm;
  if ((fuload&LR_CREATEDIBSECTION)==LR_CREATEDIBSECTION)
  { char *dbits; hbm=CreateDIBSection(NULL,(BITMAPINFO*)bih,DIB_RGB_COLORS,(void**)&dbits,NULL,0);
    unsigned int numbytes = bih->biSizeImage;
    if (numbytes==0) numbytes = ((bih->biWidth*bih->biBitCount/8+3)&0xFFFFFFFC)*bih->biHeight;
    CopyMemory(dbits,sbits,numbytes);
  }
  else
  { HDC hdc = GetDC(0);
    hbm = CreateDIBitmap(hdc,bih,CBM_INIT,sbits,(BITMAPINFO*)bih,DIB_RGB_COLORS);
    ReleaseDC(0,hdc);
  }
  delete[] buf;
  return hbm;
}

