#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>
#include <mmsystem.h>
#include "../../ogg.h"
//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#include <condefs.h>
#pragma hdrstop
USEUNIT("..\..\ogg.cpp");
#endif
//---------------------------------------------------------------------------
HINSTANCE hInstance;
HWND hmain;


void Progress(int i,unsigned long pos,unsigned long tot)
{ RECT rc; GetClientRect(hmain,&rc); int width=rc.right;
  double f = ((double)pos)/((double)tot) * ((double)width);
  int x = (int)f;
  HDC hdc = GetDC(hmain);
  rc.left=0; rc.top=100+20*i; rc.right=x; rc.bottom=rc.top+20;
  FillRect(hdc,&rc,(HBRUSH)GetStockObject(LTGRAY_BRUSH));
  rc.left=x; rc.right=width;
  FillRect(hdc,&rc,(HBRUSH)GetStockObject(WHITE_BRUSH));
  char c[200]; const char *title;
  if (i==0) title="to wav"; else title="playing";
  wsprintf(c,"%s : %lu / %lu",title,pos,tot);
  SetBkMode(hdc,TRANSPARENT);
  TextOut(hdc,20,rc.top+2,c,(int)strlen(c));
  ReleaseDC(hmain,hdc);
  // and pump messages
  MSG msg; for(;;)
  { BOOL res=PeekMessage(&msg,0,0,0,PM_REMOVE);
    if (!res||msg.message==WM_QUIT) return;
    TranslateMessage(&msg); DispatchMessage(&msg);
  }
}


// The WAV file format.
// The format is a single RIFF WAVE chunk, groupid="RIFF", rifftype="WAVE"
// This contains two chunks, first the format chunk ckID="fmt "
// and next the sound data chunk ckID="data"
// Note: all RIFF blocks contain 4byte identifier, then 4byte 'sizeofrest', then the rest of the chunk.
// The following structures represent the structure of a WAV file.
// NB. the use of pshpack1.h, to ensure that the structures don't get padded by the compiler.
#include <pshpack1.h>
typedef struct
{ char id[4];         //="fmt "
  unsigned long size; //=16
  short wFormatTag;   //=WAVE_FORMAT_PCM=1
  unsigned short wChannels;       //=1 or 2 for mono or stereo
  unsigned long dwSamplesPerSec;  //=11025 or 22050 or 44100
  unsigned long dwAvgBytesPerSec; //=wBlockAlign * dwSamplesPerSec
  unsigned short wBlockAlign;     //=wChannels * (wBitsPerSample==8?1:2)
  unsigned short wBitsPerSample;  //=8 or 16, for bits per sample
} FmtChunk;

typedef struct
{ char id[4];            //="data"
  unsigned long size;    //=datsize, size of the following array
  char data[1];          //=the raw data goes here
} DataChunk;

typedef struct
{ char id[4];         //="RIFF"
  unsigned long size; //=datsize+8+16+4
  char type[4];       //="WAVE"
  FmtChunk fmt;
  DataChunk dat;
} WavChunk;
#include <poppack.h>



// MFILE -- acts like a FILE, but for memory.
typedef struct {unsigned int pos,size; const char *buf;} MFILE;
MFILE *mopen(const char *buf,unsigned int size)
{ MFILE *m = new MFILE;
  m->size=size; m->buf=buf; m->pos=0;
  return m;
}
void mclose(MFILE *m)
{ delete m;
}
size_t mread_ogg(void *ptr,size_t size,size_t nmemb,void *src)
{ MFILE *m = (MFILE*)src; if (m==0) return 0;
  if (m->pos>m->size) return 0;
  unsigned int wanted = (unsigned int)size*nmemb;
  unsigned int actual = wanted; if (m->pos+actual>m->size) actual=m->size-m->pos;
  memcpy(ptr,m->buf+m->pos,actual);
  m->pos += actual;
  return actual;
}
int mseek_ogg(void *src,__int64 offset,int whence)
{ MFILE *m = (MFILE*)src; if (m==0) return -1;
  if (whence==SEEK_CUR) {m->pos+=(unsigned int)offset; return 0;}
  else if (whence==SEEK_END) {m->pos=m->size+(unsigned int)offset; return 0;}
  else if (whence==SEEK_SET) {m->pos=(unsigned int)offset; return 0;}
  else return -1;
}
int mclose_ogg(void *src)
{ MFILE *m = (MFILE*)src; if (m==0) return EOF;
  mclose(m); return 0;
}
long mtell_ogg(void *src)
{ MFILE *m=(MFILE*)src; if (m==0) return -1;
  return m->pos;
}
const ov_callbacks mfile_callbacks = {mread_ogg,mseek_ogg,mclose_ogg,mtell_ogg};


void OggToWav(const char *oggfn)
{ // We could easily load the ogg directly from a file.
  // But instead we're going to load the ogg from a memory block.
  // So first set up the memory block
  HANDLE hoggf = CreateFile(oggfn,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
  if (hoggf==INVALID_HANDLE_VALUE) return;
  DWORD size = GetFileSize(hoggf,NULL);
  char *oggbuf = new char[size];
  DWORD red; ReadFile(hoggf,oggbuf,size,&red,NULL);
  CloseHandle(hoggf);
  // now point ogg to it
  OggVorbis_File vf;
  MFILE *min = mopen(oggbuf,size);
  if (min==NULL) {delete[] oggbuf; return;}
  int res=ov_open_callbacks(min,&vf,NULL,0,mfile_callbacks);
  if (res<0) {mclose(min); delete[] oggbuf; return;}
  // note that, if ov_open succeeds, then vf has ownership of the file (be it
  // disk or resource) and will be responsible for closing it.
  vorbis_info *vi=ov_info(&vf,-1);
  if (!ov_seekable(&vf)) {ov_clear(&vf); delete[] oggbuf; return;}
  //
  // Now we will create a WAV format laid out in memory.
  // It's possible (and easy!) to play this directly, using
  // the call PlaySound(). But instead, we're just going to
  // write it to disk.
  //
  // Problem: although ogg claims to give us ov_pcm_total, it shortchanges
  // us a bit. Hence we leave some extra safety margin here, and reset
  // the header at the end
  const unsigned int extra=10000;
  unsigned int numsamples = extra+(unsigned int)ov_pcm_total(&vf,0);
  unsigned int numbytes = numsamples*vi->channels*2; // 16bit output wav
  unsigned int wavsize = sizeof(WavChunk)+numbytes;
  char *wav = new char[wavsize]; WavChunk *wc = (WavChunk*)wav;
  // this following is just the layout of a WAV file.
  strncpy(wc->id,"RIFF",4);
  wc->size = 4 + sizeof(FmtChunk)+sizeof(DataChunk)+numbytes;
  strncpy(wc->type,"WAVE",4);
  strncpy(wc->fmt.id,"fmt ",4);
  wc->fmt.size=16;
  wc->fmt.wFormatTag=1;
  wc->fmt.wChannels = (unsigned short)vi->channels;
  wc->fmt.dwSamplesPerSec = vi->rate;
  wc->fmt.wBitsPerSample = 16;
  wc->fmt.wBlockAlign = (unsigned short)(wc->fmt.wChannels * wc->fmt.wBitsPerSample/8);
  wc->fmt.dwAvgBytesPerSec = wc->fmt.wBlockAlign * wc->fmt.dwSamplesPerSec;
  strncpy(wc->dat.id,"data",4);
  wc->dat.size = numbytes;
  char *buf = wc->dat.data; unsigned long bufpos=0;

  // now do the actual translating of raw data
  int current_section=0; for(;;)
  { long ret = ov_read(&vf,buf+bufpos,numbytes-bufpos,0,2,1,&current_section);
    if (ret<=0) break;
    else bufpos+=ret;
    Progress(0,bufpos,numbytes);
  }

  // now recalculate the actual size
  numbytes = bufpos;
  //numsamples = numbytes/2/vi->channels;
  wc->dat.size = numbytes;

  ov_clear(&vf);
  delete[] oggbuf;

  // Now: we have allocated wav[wavsize]
  // of which the first sizeof(WavChunk)+numbytes are the actual WAV file
  // and the rest are junk. And we have numsamples, giving a playtime
  // of milliseconds 1000*numsamples/(unsigned int)wc->fmt.dwSamplesPerSec
  // It'd be really easy now to play it: just
  //   PlaySound((LPCTSTR)wav,NULL,SND_MEMORY|SND_ASYNC)
  // But instead we'll save it to disk.


  // first set up the filenames
  char wavfn[MAX_PATH]; strcpy(wavfn,oggfn);
  char *c=strrchr(wavfn,'.');
  if (c!=0 && stricmp(c,".ogg")==0) strcpy(c,".wav");
  else strcat(wavfn,".wav");
  //
  HANDLE hwavf = CreateFile(wavfn,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
  if (hwavf==INVALID_HANDLE_VALUE) MessageBox(NULL,"Error writing wav file",wavfn,MB_OK);
  else
  { DWORD writ; WriteFile(hwavf,wav,sizeof(WavChunk)+numbytes,&writ,NULL);
    CloseHandle(hwavf);
  }
  delete[] wav;
}


// For playing, we need to protect our data against concurrency.
// We store nused, the number of used buffers. This is protected by critsec.
// Every time a buffer is freed, the event huponfree is triggered.
class TBufferInfo
{ public:
  int nused; CRITICAL_SECTION critsec; HANDLE huponfree;
  TBufferInfo() {nused=0; InitializeCriticalSection(&critsec); huponfree=CreateEvent(NULL,FALSE,FALSE,NULL);}
  ~TBufferInfo() {DeleteCriticalSection(&critsec); CloseHandle(huponfree);}
};

void CALLBACK PlayCallback(HWAVEOUT, UINT msg, DWORD param, DWORD, DWORD)
{ if (msg!=WOM_DONE) return;
  // invariant: if a buffer is being freed, then nused>0.
  TBufferInfo *bufferinfo = (TBufferInfo*)(DWORD_PTR)param;
  EnterCriticalSection(&bufferinfo->critsec);
  bufferinfo->nused--;
  LeaveCriticalSection(&bufferinfo->critsec);
  SetEvent(bufferinfo->huponfree);
}

DWORD WINAPI Play(void *param)
{ // For small sounds, it'd be easier to uncompress the entire ogg
  // to a WAV structure in memory and then play it using PlaySound.
  // But that's not sensible for larger sounds. So here we'll
  // decompress it one block at a time, playing each block using
  // the multimedia api.
  const char *oggfn = (const char*)param;
  OggVorbis_File vf;
  FILE *fin = fopen(oggfn,"rb"); if (fin==NULL) return 0;
  int res=ov_open(fin,&vf,NULL,0);
  if (res<0) {fclose(fin); return 0;}
  if (!ov_seekable(&vf)) {ov_clear(&vf); return 0;}
  vorbis_info *vi=ov_info(&vf,-1);

  // set up our buffer counting scheme
  const int nbuffers = 5;
  const int bufsize = 4096;
  int currentbuffer=0;
  TBufferInfo bufferinfo;
  // open the waveout device
  WAVEFORMATEX wfx;
  wfx.wFormatTag      = WAVE_FORMAT_PCM;
  wfx.nChannels       = (WORD)vi->channels;
  wfx.wBitsPerSample  = 16;
  wfx.nBlockAlign     = (WORD)((wfx.wBitsPerSample>>3)*wfx.nChannels);
  wfx.nSamplesPerSec  = vi->rate;
  wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
  unsigned int numsamples = (unsigned int)ov_pcm_total(&vf,0);
  unsigned int numbytes = numsamples*vi->channels*wfx.wBitsPerSample/2;
  HWAVEOUT hwave;
  MMRESULT mres = waveOutOpen(&hwave,WAVE_MAPPER,&wfx,(DWORD_PTR)PlayCallback,(DWORD_PTR)&bufferinfo,CALLBACK_FUNCTION);
  if (mres!=MMSYSERR_NOERROR) return 0;
  // set up our buffers
  WAVEHDR hdrs[nbuffers];
  for (int i=0; i<nbuffers; i++)
  { ZeroMemory(&hdrs[i],sizeof(WAVEHDR));
    hdrs[i].lpData = (LPSTR)LocalAlloc(LMEM_FIXED, bufsize);
    hdrs[i].dwBufferLength=bufsize;
    waveOutPrepareHeader(hwave, &hdrs[i], sizeof(WAVEHDR));
  }

  unsigned int totprogress=0;
  int current_section=0; for (;;)
  { // wait until at least one buffer is free. Invariant:
    // if at least one buffer is free, then currentbuffer is free.
    for (bool okay=false; !okay; )
    { EnterCriticalSection(&bufferinfo.critsec);
      if (bufferinfo.nused<nbuffers) {bufferinfo.nused++; okay=true;}
      LeaveCriticalSection(&bufferinfo.critsec);
      if (!okay) WaitForSingleObject(bufferinfo.huponfree,INFINITE);
    }
    // get the data into the current buffer
    WAVEHDR &hdr = hdrs[currentbuffer];
    for (hdr.dwBufferLength=0; hdr.dwBufferLength<bufsize; )
    { long ret = ov_read(&vf,hdr.lpData+hdr.dwBufferLength,bufsize-hdr.dwBufferLength, 0,2,1,&current_section);
      if (ret<=0) break;
      hdr.dwBufferLength += ret;
      totprogress+=hdr.dwBufferLength;
      Progress(1,totprogress,numbytes);
    }
    currentbuffer = (currentbuffer+1)%nbuffers;
    // and set it playing!
    waveOutWrite(hwave, &hdr, sizeof(WAVEHDR));
    if (hdr.dwBufferLength==0) break;
  }

  // wait for it to finish
  for (bool okay=false; !okay; )
  { EnterCriticalSection(&bufferinfo.critsec);
    okay = (bufferinfo.nused==0);
    LeaveCriticalSection(&bufferinfo.critsec);
    if (!okay) WaitForSingleObject(bufferinfo.huponfree,INFINITE);
  }
  // cleanup
  waveOutReset(hwave);
  for (int i=0; i<nbuffers; i++)
  { waveOutUnprepareHeader(hwave, &hdrs[i], sizeof(WAVEHDR));
    LocalFree(hdrs[i].lpData);
  }
  waveOutClose(hwave);
  ov_clear(&vf);
  return 0;
}

bool isbusy=false;

LRESULT CALLBACK PlainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{ switch (msg)
  { case WM_CREATE:
    { DragAcceptFiles(hwnd,TRUE);
    } break;
    case WM_PAINT:
    { PAINTSTRUCT ps; BeginPaint(hwnd,&ps); const char *c;
      c="Drag an OGG file onto this window."; TextOut(ps.hdc,20,20,c,(int)strlen(c));
      c="It will be played,"; TextOut(ps.hdc,20,40,c,(int)strlen(c));
      c="and also converted into a WAV."; TextOut(ps.hdc,20,60,c,(int)strlen(c));
      EndPaint(hwnd,&ps);
    } break;
    case WM_DROPFILES:
    { SetCursor(LoadCursor(NULL,IDC_WAIT));
      isbusy=true;
      HDROP hdrop = (HDROP)wParam;
      int num = DragQueryFile(hdrop,0xFFFFFFFF,NULL,0);
      for (int i=0; i<num; i++)
      { int len = DragQueryFile(hdrop,i,0,0);
        char *c=new char[len+1];
        DragQueryFile(hdrop,i,c,len+1);
        // We'll start it playing in a separate thread. This way,
        // the user will hear the music while we're waiting for the
        // OggToWav to complete. We put the thread at a higher priority
        // because audio sounds terrible if it's broken.
        DWORD tid; HANDLE hthread=CreateThread(NULL,0,Play,c,0,&tid);
        SetThreadPriority(hthread,THREAD_PRIORITY_ABOVE_NORMAL);
        OggToWav(c); MessageBeep(0); // beep when the wav diskfile is finished
        WaitForSingleObject(hthread,INFINITE); CloseHandle(hthread);
        delete[] c;
      }
      DragFinish(hdrop);
      isbusy=false;
      SetCursor(LoadCursor(NULL,IDC_ARROW));
    } break;
    case WM_CLOSE:
    { if (isbusy) {MessageBeep(0); return 0;}
      DestroyWindow(hwnd);
    } break;
    case WM_DESTROY:
    { DragAcceptFiles(hwnd,FALSE);
      PostQuitMessage(0);
    } break;
  }
  return DefWindowProc(hwnd, msg, wParam, lParam);
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int showcmd)
{ hInstance=h;
  //
  WNDCLASSEX wcex; ZeroMemory(&wcex,sizeof(wcex)); wcex.cbSize = sizeof(WNDCLASSEX);
  BOOL res=GetClassInfoEx(hInstance,"PlainClass",&wcex);
  if (!res)
  { wcex.style = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc = (WNDPROC)PlainWndProc;
    wcex.cbClsExtra = 0;
    wcex.cbWndExtra = 0;
    wcex.hInstance = hInstance;
    wcex.hIcon = NULL;
    wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wcex.lpszMenuName	= NULL;
    wcex.lpszClassName = "PlainClass";
    wcex.hIconSm = NULL;
    ATOM res=RegisterClassEx(&wcex);
    if (res==0) {MessageBox(NULL,"Failed to register class","Error",MB_OK); return 0;}
  }
  //
  hmain = CreateWindowEx(0,"PlainClass", "Ogg Test", WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT, 400, 400, NULL, NULL, hInstance, NULL);
  if (hmain==NULL) {MessageBox(NULL,"Failed to create window","Error",MB_OK); return 0;}
  ShowWindow(hmain,showcmd);
  //
  MSG msg;
  while (GetMessage(&msg, NULL, 0, 0))
  { TranslateMessage(&msg);
	DispatchMessage(&msg);
  }
  return (int)msg.wParam;
}

 
