#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commdlg.h>
#include <string>
#include <list>
using namespace std;
#include "../../unzip.h"
#include "../../zip.h"
//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#include <condefs.h>
#pragma hdrstop
USERC("menu.rc");
USEUNIT("..\..\unzip.cpp");
USEUNIT("..\..\zip.cpp");
#endif
//---------------------------------------------------------------------------
HINSTANCE hInstance;
HWND hmain,hlbs,hed;
typedef struct {string name; const char *buf; unsigned int size; bool isbmp;} TBlob;
list<TBlob> blobs;
int sel;
char fn[MAX_PATH];

void Redraw(bool dolist=true)
{ if (dolist)
  { int oldsel=sel; sel=-1;
    SendMessage(hlbs,LB_RESETCONTENT,0,0);
    for (list<TBlob>::iterator i=blobs.begin(); i!=blobs.end(); i++)
    { const char *c = i->name.c_str();
      SendMessage(hlbs,LB_ADDSTRING,0,(LPARAM)c);
    }
    if (oldsel>=(int)blobs.size()) oldsel=-1;
    SendMessage(hlbs,LB_SETSEL,TRUE,oldsel);
    sel=oldsel;
  }
  if (sel==-1) ShowWindow(hed,SW_HIDE);
  else
  { list<TBlob>::const_iterator ci=blobs.begin();
    for (int i=0; i<sel; i++) ci++;
    if (ci->isbmp) ShowWindow(hed,SW_HIDE);
    else
    { ShowWindow(hed,SW_SHOW);
      SendMessage(hed,WM_SETTEXT,0,(LPARAM)ci->buf);
    }
  }
  InvalidateRect(hmain,NULL,TRUE);
}

void FileNew()
{ for (list<TBlob>::iterator i=blobs.begin(); i!=blobs.end(); i++)
  { delete[] i->buf;
  }
  blobs.clear(); sel=-1; *fn=0; Redraw();
}

void FileOpen()
{ char tfn[MAX_PATH]; *tfn=0;
  OPENFILENAME ofn; ZeroMemory(&ofn,sizeof(ofn));
  ofn.lStructSize=sizeof(ofn);
  ofn.hwndOwner=hmain;
  ofn.lpstrFilter="ZFormat documents\0*.zfd\0";
  ofn.lpstrCustomFilter=NULL;
  ofn.nFilterIndex=0;
  ofn.lpstrFile=tfn;
  ofn.nMaxFile=MAX_PATH;
  ofn.lpstrFileTitle=NULL;
  ofn.lpstrInitialDir=NULL;
  ofn.lpstrTitle=NULL;
  ofn.lpstrDefExt = "zfd";
  ofn.Flags=OFN_FILEMUSTEXIST;
  BOOL res = GetOpenFileName(&ofn);
  if (!res) return;
  FileNew();
  // read the header
  HANDLE hf=CreateFile(tfn,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
  bool okay=false;
  if (hf!=INVALID_HANDLE_VALUE)
  { char magic[5]; DWORD red;
    ReadFile(hf,magic,4,&red,NULL); magic[4]=0;
    if (strcmp(magic,"ZFOR")==0)
    { for (;;)
      { *magic=1; ReadFile(hf,magic,1,&red,NULL);
        if (*magic==0 || red==0) break;
      }
      if (red==1 && *magic==0)
      { okay=true;
      }
    }
  }
  if (!okay && hf==INVALID_HANDLE_VALUE) CloseHandle(hf);
  if (!okay) MessageBox(hmain,"Error reading document","ZFormat",MB_OK);
  if (!okay) return;
  SetCapture(hmain);
  SetCursor(LoadCursor(NULL,IDC_WAIT));
  // read the data
  HZIP hz = OpenZip(hf,0,ZIP_HANDLE);
  ZIPENTRY ze; GetZipItem(hz,-1,&ze); int numentries=ze.index;
  for (int i=0; i<numentries; i++)
  { GetZipItem(hz,i,&ze);
    char *buf = new char[ze.unc_size];
    UnzipItem(hz,i,buf,ze.unc_size,ZIP_MEMORY);
    TBlob b;
    b.buf=buf; b.size=ze.unc_size; b.name=string(ze.name+1); b.isbmp=(ze.name[0]=='B');
    blobs.push_back(b);
  }
  CloseZip(hz);
  CloseHandle(hf);
  SetCursor(LoadCursor(NULL,IDC_ARROW));
  ReleaseCapture();
  sel=-1;
  Redraw();
}

void FileSaveAs()
{ char tfn[MAX_PATH]; *tfn=0;
  OPENFILENAME ofn; ZeroMemory(&ofn,sizeof(ofn));
  ofn.lStructSize=sizeof(ofn);
  ofn.hwndOwner=hmain;
  ofn.lpstrFilter="ZFormat documents\0*.zfd\0";
  ofn.lpstrCustomFilter=NULL;
  ofn.nFilterIndex=0;
  ofn.lpstrFile=tfn;
  ofn.nMaxFile=MAX_PATH;
  ofn.lpstrFileTitle=NULL;
  ofn.lpstrInitialDir=NULL;
  ofn.lpstrTitle=NULL;
  ofn.lpstrDefExt="zfd";
  ofn.Flags=OFN_OVERWRITEPROMPT|OFN_PATHMUSTEXIST;
  BOOL res = GetSaveFileName(&ofn);
  if (!res) return;
  // write the header
  HANDLE hf=CreateFile(tfn,GENERIC_WRITE,FILE_SHARE_READ,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
  if (hf==INVALID_HANDLE_VALUE) {MessageBox(hmain,"Error writing document","ZFormat",MB_OK); return;}
  SetCapture(hmain);
  SetCursor(LoadCursor(NULL,IDC_WAIT));
  const char *c = "ZFOR\r\nThe first few lines are text in this file format, but the rest is zipped up data\r\n";
  DWORD writ; WriteFile(hf,c,(DWORD)strlen(c)+1,&writ,NULL);
  // write the data
  HZIP hz = CreateZip(hf,0,ZIP_HANDLE);
  for (list<TBlob>::const_iterator ci=blobs.begin(); ci!=blobs.end(); ci++)
  { string bzn; if (ci->isbmp) bzn="B"+ci->name; else bzn="T"+ci->name;
    ZipAdd(hz,bzn.c_str(), (void*)ci->buf,ci->size,ZIP_MEMORY);
  }
  CloseZip(hz);
  CloseHandle(hf);
  SetCursor(LoadCursor(NULL,IDC_ARROW));
  ReleaseCapture();
}

void Insert(bool istext)
{ char tfn[MAX_PATH]; *tfn=0;
  OPENFILENAME ofn; ZeroMemory(&ofn,sizeof(ofn));
  ofn.lStructSize=sizeof(ofn);
  ofn.hwndOwner=hmain;
  ofn.lpstrFilter= (istext?"Text files\0*.txt\0":"Bitmaps\0*.bmp\0");
  ofn.lpstrCustomFilter=NULL;
  ofn.nFilterIndex=0;
  ofn.lpstrFile=tfn;
  ofn.nMaxFile=MAX_PATH;
  ofn.lpstrFileTitle=NULL;
  ofn.lpstrInitialDir=NULL;
  ofn.lpstrTitle=NULL;
  ofn.Flags=OFN_FILEMUSTEXIST;
  ofn.lpstrDefExt = (istext?"txt":"bmp");
  BOOL res = GetOpenFileName(&ofn);
  if (!res) return;
  // read in the raw data
  HANDLE hf = CreateFile(tfn,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
  if (hf==INVALID_HANDLE_VALUE) return;
  unsigned int size=GetFileSize(hf,NULL);
  char *buf = new char[size+1];
  DWORD red; ReadFile(hf,buf,size,&red,NULL);
  CloseHandle(hf);
  buf[size]=0;
  // pick our internal name for it
  string newname=string(tfn+ofn.nFileOffset);
  string tname;
  for (char c='a'; ; c++)
  { if (c=='a') tname=newname; else tname=string(&c,1)+newname;
    bool isin=false;
    for (list<TBlob>::const_iterator i=blobs.begin(); i!=blobs.end(); i++)
    { if (stricmp(i->name.c_str(), tname.c_str())==0) isin=true;
    }
    if (!isin) break;
    if (c=='z') return;
  }
  // and add it to the list
  TBlob b; b.buf=buf; b.size=size; b.isbmp=!istext; b.name=tname;
  blobs.push_back(b);
  sel=(int)blobs.size()-1;
  Redraw();
}

void Delete()
{ sel=(int)SendMessage(hlbs,LB_GETCURSEL,0,0);
  if (sel==LB_ERR) {sel=-1; return;}
  list<TBlob>::iterator ci=blobs.begin(); for (int i=0; i<sel; i++) ci++;
  delete ci->buf;
  blobs.erase(ci);
  Redraw(); 
}

HBITMAP LoadMemImage(HINSTANCE,LPCTSTR buf,UINT uType,int cxDesired,int cyDesired,UINT fuLoad)
{ if (buf==0 || uType!=IMAGE_BITMAP || cxDesired!=0 || cyDesired!=0 || fuLoad!=LR_CREATEDIBSECTION) return 0;
  BITMAPFILEHEADER *bfh=(BITMAPFILEHEADER*)buf;
  BITMAPINFOHEADER *bih=(BITMAPINFOHEADER*)(buf+sizeof(BITMAPFILEHEADER));
  //RGBQUAD *cols = (RGBQUAD*)(buf+sizeof(BITMAPFILEHEADER)+bih->biSize);
  char *sbits = (char*)(buf+bfh->bfOffBits);
  char *dbits; HBITMAP hbm=CreateDIBSection(NULL,(BITMAPINFO*)bih,DIB_RGB_COLORS,(void**)&dbits,NULL,0);
  unsigned int numbytes=bih->biSizeImage;
  if (numbytes==0) numbytes = ((bih->biWidth*bih->biBitCount/8+3)&0xFFFFFFFC)*bih->biHeight;
  CopyMemory(dbits,sbits,numbytes);
  return hbm;
}


LRESULT CALLBACK PlainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{ switch (msg)
  { case WM_CREATE:
    { hlbs = CreateWindow("LISTBOX","",WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS|LBS_DISABLENOSCROLL|LBS_HASSTRINGS|LBS_NOINTEGRALHEIGHT|LBS_NOTIFY|WS_VSCROLL|WS_BORDER,
          0,0,0,0,hwnd,(HMENU)1,hInstance,NULL);
      hed = CreateWindow("EDIT","",WS_CHILD|WS_VISIBLE|WS_VSCROLL|WS_HSCROLL|ES_AUTOHSCROLL|ES_AUTOVSCROLL|ES_MULTILINE|ES_READONLY,
               0,0,0,0,hwnd,(HMENU)1,hInstance,NULL);
      sel=-1;
    } break;
    case WM_SIZE:
    { int width=LOWORD(lParam),height=HIWORD(lParam);
      MoveWindow(hlbs,0,0,100,height,TRUE);
      MoveWindow(hed,100,0,width-100,height,TRUE);
    }
    case WM_PAINT:
    { PAINTSTRUCT ps; BeginPaint(hwnd,&ps);
      FillRect(ps.hdc,&ps.rcPaint,(HBRUSH)GetStockObject(GRAY_BRUSH));
      RECT rc; GetClientRect(hwnd,&rc); int width=rc.right,height=rc.bottom;
      if (sel!=-1)
      { list<TBlob>::const_iterator ci=blobs.begin();
        for (int i=0; i<sel; i++) ci++;
        if (ci->isbmp)
        { HBITMAP hbm = LoadMemImage(hInstance,ci->buf,IMAGE_BITMAP,0,0,LR_CREATEDIBSECTION);
          HDC sdc=GetDC(0); HDC tdc = CreateCompatibleDC(sdc); ReleaseDC(0,sdc);
          HGDIOBJ holdbm = SelectObject(tdc,hbm);
          BITMAP bmp; GetObject(hbm,sizeof(bmp),&bmp);
          SetStretchBltMode(ps.hdc,COLORONCOLOR);
          StretchBlt(ps.hdc,100,0,width-100,height,tdc,0,0,bmp.bmWidth,bmp.bmHeight,SRCCOPY);
          SelectObject(tdc,holdbm);
          DeleteDC(tdc);
          DeleteObject(hbm);
        }
      }
      EndPaint(hwnd,&ps);
    } break;
    case WM_COMMAND:
    { int id=LOWORD(wParam), code=HIWORD(wParam);
      if (id==101) FileNew();
      else if (id==102) FileOpen();
      else if (id==103) FileSaveAs();
      else if (id==104) SendMessage(hwnd,WM_CLOSE,0,0);
      else if (id==201) Insert(true);
      else if (id==202) Insert(false);
      else if (id==203) Delete();
      else if (id==1 && code==LBN_SELCHANGE) {sel=(int)SendMessage(hlbs,LB_GETCURSEL,0,0); Redraw(false);}
    } break;
    case WM_CLOSE:
    { DestroyWindow(hwnd);
    } break;
    case WM_DESTROY:
    { FileNew();
      PostQuitMessage(0);
    } break;
  }
  return DefWindowProc(hwnd, msg, wParam, lParam);
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int showcmd)
{ hInstance=h;
  //
  WNDCLASSEX wcex; ZeroMemory(&wcex,sizeof(wcex)); wcex.cbSize = sizeof(WNDCLASSEX);
  BOOL res=GetClassInfoEx(hInstance,"PlainClass",&wcex);
  if (!res)
  { wcex.style = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc = (WNDPROC)PlainWndProc;
    wcex.cbClsExtra = 0;
    wcex.cbWndExtra = 0;
    wcex.hInstance = hInstance;
    wcex.hIcon = NULL;
    wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wcex.lpszMenuName	= "MAINMENU";
    wcex.lpszClassName = "PlainClass";
    wcex.hIconSm = NULL;
    ATOM res=RegisterClassEx(&wcex);
    if (res==0) {MessageBox(NULL,"Failed to register class","Error",MB_OK); return 0;}
  }
  //
  hmain = CreateWindowEx(0,"PlainClass", "ZFormat Compound Document", WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT, 400, 400, NULL, NULL, hInstance, NULL);
  if (hmain==NULL) {MessageBox(NULL,"Failed to create window","Error",MB_OK); return 0;}
  ShowWindow(hmain,showcmd);
  //
  MSG msg;
  while (GetMessage(&msg, NULL, 0, 0))
  { TranslateMessage(&msg);
	DispatchMessage(&msg);
  }
  return (int)msg.wParam;
}


