#define STRICT
#define NO_WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shlobj.h>
#include "../../zip.h"
//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#include <condefs.h>
#pragma hdrstop
USEUNIT("..\..\zip.cpp");
#endif
//---------------------------------------------------------------------------
HINSTANCE hInstance;
HWND hed;

HZIP hz=0;
char hzfn[MAX_PATH]={0};

void log(const char *c)
{ char *d = new char[strlen(c)+4];
  wsprintf(d,"%s\r\n",c);
  SendMessage(hed,EM_SETSEL,(WPARAM)-1,-1);
  SendMessage(hed,EM_REPLACESEL,FALSE,(LPARAM)d);
  delete[] d;
}

// we will pick a name for the zipfile being created
void PickName()
{ IMalloc *shmalloc; SHGetMalloc(&shmalloc);
  ITEMIDLIST *pidl; SHGetSpecialFolderLocation(NULL,CSIDL_DESKTOP,&pidl);
  char dpath[MAX_PATH]; SHGetPathFromIDList(pidl,dpath);
  shmalloc->Free(pidl);
  shmalloc->Release();
  for (int i=0; ; i++)
  { wsprintf(hzfn,"%s\\zipdrop-%i.zip",dpath,i);
    DWORD attr = GetFileAttributes(hzfn);
    if (attr==0xFFFFFFFF) return;
  }
}


void RecAddStuff(HZIP hz,const char *fn,const char *dir)
{ // first, get the name of this item itself
  const char *zfn=strrchr(fn,'/');
  if (zfn==0) zfn=strrchr(fn,'\\');
  if (zfn==0) zfn=fn; else zfn++;
  // make 'zp' be the path+name of the entity inside the zip file
  char zp[MAX_PATH];
  if (dir==0 || *dir==0) strcpy(zp,zfn);
  else wsprintf(zp,"%s/%s",dir,zfn);
  // now figure out what kind of thing it is
  DWORD attr = GetFileAttributes(fn);
  if ((attr&FILE_ATTRIBUTE_DIRECTORY)==0)
  { ZipAdd(hz, zp, (char*)fn,0,ZIP_FILENAME);
    log(zp);
    return;
  }
  // otherwise, it's a directory, so we have to recursively add it
  ZipAdd(hz,zp, 0,0,ZIP_FOLDER);
  log(zp);
  char mask[MAX_PATH]; wsprintf(mask,"%s\\*",fn);
  WIN32_FIND_DATA fdat;
  HANDLE hfind = FindFirstFile(mask,&fdat);
  while (hfind!=INVALID_HANDLE_VALUE)
  { bool issilly = strcmp(fdat.cFileName,".")==0 || strcmp(fdat.cFileName,"..")==0;
    if (!issilly)
    { char subfn[MAX_PATH]; wsprintf(subfn,"%s\\%s",fn,fdat.cFileName);
      RecAddStuff(hz,subfn,zp);
    }
    BOOL res = FindNextFile(hfind,&fdat);
    if (!res) {FindClose(hfind); hfind=INVALID_HANDLE_VALUE;}
  }
}


LRESULT CALLBACK PlainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{ switch (msg)
  { case WM_CREATE:
    { DragAcceptFiles(hwnd,TRUE);
      hed = CreateWindow("EDIT","",WS_CHILD|WS_VISIBLE|WS_VSCROLL|WS_HSCROLL|ES_AUTOHSCROLL|ES_AUTOVSCROLL|ES_MULTILINE|ES_READONLY,
               0,0,0,0,hwnd,(HMENU)1,hInstance,NULL);
      log("This program will create a zip file.");
      log("Drag files onto this window to zip them.");

    } break;
    case WM_SIZE:
    { int width=LOWORD(lParam), height=HIWORD(lParam);
      MoveWindow(hed,0,0,width,height,TRUE);
    }
    case WM_PAINT:
    { PAINTSTRUCT ps; BeginPaint(hwnd,&ps); EndPaint(hwnd,&ps);
    } break;
    case WM_DROPFILES:
    { SetCursor(LoadCursor(NULL,IDC_WAIT));
      if (hz==0)
      { PickName(); hz=CreateZip(hzfn,0,ZIP_FILENAME);
        log("----------"); log("The zip file will be called:");
        log(hzfn); log("----------");
      }
      HDROP hdrop = (HDROP)wParam;
      int num = DragQueryFile(hdrop,0xFFFFFFFF,NULL,0);
      for (int i=0; i<num; i++)
      { int len = DragQueryFile(hdrop,i,0,0);
        char *c=new char[len+1];
        DragQueryFile(hdrop,i,c,len+1);
        RecAddStuff(hz,c,"");
        delete[] c;
      }
      DragFinish(hdrop);
      SetCursor(LoadCursor(NULL,IDC_ARROW));
    } break;
    case WM_CLOSE:
    { if (hz!=0) CloseZip(hz); hz=0;
      DestroyWindow(hwnd);
    } break;
    case WM_DESTROY:
    { DragAcceptFiles(hwnd,FALSE);
      PostQuitMessage(0);
    } break;
  }
  return DefWindowProc(hwnd, msg, wParam, lParam);
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ hInstance=h;
  //
  //
  WNDCLASSEX wcex; ZeroMemory(&wcex,sizeof(wcex)); wcex.cbSize = sizeof(WNDCLASSEX);
  BOOL res=GetClassInfoEx(hInstance,"PlainClass",&wcex);
  if (!res)
  { wcex.style = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc = (WNDPROC)PlainWndProc;
    wcex.cbClsExtra = 0;
    wcex.cbWndExtra = 0;
    wcex.hInstance = hInstance;
    wcex.hIcon = NULL;
    wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wcex.lpszMenuName	= NULL;
    wcex.lpszClassName = "PlainClass";
    wcex.hIconSm = NULL;
    ATOM res=RegisterClassEx(&wcex);
    if (res==0) {MessageBox(NULL,"Failed to register class","Error",MB_OK); return 0;}
  }
  //
  HWND hwnd = CreateWindowEx(0,"PlainClass", "Zip Drop", WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT, 500, 200, NULL, NULL, hInstance, NULL);
  if (hwnd==NULL) {MessageBox(NULL,"Failed to create window","Error",MB_OK); return 0;}
  ShowWindow(hwnd,SW_SHOW);
  //
  MSG msg;
  while (GetMessage(&msg, NULL, 0, 0))
  { TranslateMessage(&msg);
	DispatchMessage(&msg);
  }
  return (int)msg.wParam;
}




