#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "../../zip.h"
#include "../../unzip.h"
//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#include <condefs.h>
USEUNIT("..\..\unzip.cpp");
USEUNIT("..\..\zip.cpp");
#endif
//---------------------------------------------------------------------------
HINSTANCE hInstance;


DWORD CALLBACK WriterFunc(void *dat)
{ HANDLE hwrite = (HANDLE)dat;
  char buf[1000];
  for (int i=0; i<1000; i++) buf[i] = (char)((1000-i)&0x7F);
  DWORD writ; WriteFile(hwrite,buf,1000,&writ,NULL);
  // the call to WriteFile call blocks until all bytes have been written
  CloseHandle(hwrite);
  return 0;
}

DWORD CALLBACK ReaderFunc(void *dat)
{ HANDLE hread = (HANDLE)dat;
  HANDLE hf = CreateFile("ziptest-topipe.zip",GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
  char buf[1000];
  for (;;)
  { DWORD red; ReadFile(hread,buf,1000,&red,NULL);
    if (red==0) break; // reached the end!
    DWORD writ; WriteFile(hf,buf,red,&writ,NULL);
  }
  CloseHandle(hf);
  return 0;
}

DWORD CALLBACK ReaderFunc2(void *dat)
{ HANDLE hread = (HANDLE)dat;
  HANDLE hf = CreateFile("unztest-topipe.zip",GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
  char buf[1000];
  for (;;)
  { DWORD red; ReadFile(hread,buf,1000,&red,NULL);
    if (red==0) break; // reached the end!
    DWORD writ; WriteFile(hf,buf,red,&writ,NULL);
  }
  CloseHandle(hf);
  return 0;
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ hInstance=h;
  HANDLE hf,hread,hwrite,hthread; HZIP hz; DWORD writ;

  char path[MAX_PATH]; GetModuleFileName(hInstance,path,MAX_PATH);
  char *c=strrchr(path,'/'); if (c==0) c=strrchr(path,'\\'); if (c==0) c=path;
  strcpy(c,"\\ziptest-temp-files"); CreateDirectory(path,NULL);
  SetCurrentDirectory(path);

  // first we'll create a source file to work with
  hf = CreateFile("ziptest-source.dat",GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
  if (hf==NULL) return MessageBox(NULL,"Unable to create temporary source file","Error",MB_OK);
  char buf[1000]; for (int i=0; i<1000; i++) buf[i] = (char)(i&0x7F);
  WriteFile(hf,buf,1000,&writ,NULL);
  CloseHandle(hf);


  MessageBox(NULL,"Testing zipfile creation...","Ziptest",MB_OK);


  // (1) create a zip file directly
  hz = CreateZip("ziptest-tofile.zip",0,ZIP_FILENAME);
  ZipAdd(hz,"tof-1.dat",  "ziptest-source.dat",0,ZIP_FILENAME);
  CloseZip(hz);


  // (2) create a zip to a file handle
  hf = CreateFile("ziptest-tofilehandle.zip",GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
  hz = CreateZip(hf,0,ZIP_HANDLE);
  CloseHandle(hf);
  ZipAdd(hz,"tofh-1.dat",  "ziptest-source.dat",0,ZIP_FILENAME);
  CloseZip(hz);

  // (3) create a zip to a system pagefile memory block
  hz = CreateZip(0,100000, ZIP_MEMORY);
  ZipAdd(hz,"tom-1.dat",  "ziptest-source.dat",0,ZIP_FILENAME);
  void *buf2; unsigned long len; ZipGetMemory(hz, &buf2, &len);
  hf = CreateFile("ziptest-topagemem.zip",GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
  WriteFile(hf,buf2,len,&writ,NULL);
  CloseHandle(hf);
  CloseZip(hz);

  // (4) create a zip file to our own memory block
  char buf3[10000];
  hz = CreateZip(buf3,10000, ZIP_MEMORY);
  ZipAdd(hz,"tom-1.dat",  "ziptest-source.dat",0,ZIP_FILENAME);
  ZipGetMemory(hz, NULL, &len);
  hf = CreateFile("ziptest-tomem.zip",GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
  WriteFile(hf,buf3,len,&writ,NULL);
  CloseHandle(hf);
  CloseZip(hz);


  // (5) create a zip file to a pipe
  CreatePipe(&hread,&hwrite,NULL,0);
  unsigned long tid; hthread = CreateThread(NULL,0,ReaderFunc,(void*)hread,0,&tid);
  hz = CreateZip(hwrite,0, ZIP_HANDLE);
  ZipAdd(hz,"toph-1.dat",  "ziptest-source.dat",0,ZIP_FILENAME);
  CloseZip(hz);
  CloseHandle(hwrite);
  WaitForSingleObject(hthread,INFINITE);
  CloseHandle(hthread);
  CloseHandle(hread);



  hz = CreateZip("ziptest-fromstuff.zip",0,ZIP_FILENAME);

  // (6) add a folder
  ZipAdd(hz,"subfolder",  0,0,ZIP_FOLDER);

  // (7) add from a file into a folder
  ZipAdd(hz,"subfolder\\from-file.dat",   "ziptest-source.dat",0,ZIP_FILENAME);
  ZipAdd(hz,"otherfolder\\from-file.dat", "ziptest-source.dat",0,ZIP_FILENAME);

  // (8) add from a file handle
  hf = CreateFile("ziptest-source.dat",GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
  ZipAdd(hz,"from-file-handle.dat", hf,0,ZIP_HANDLE);
  CloseHandle(hf);

  // (9) add from a memory block
  char buf4[1000];
  for (int i=0; i<1000; i++) buf4[i] = (char)(i&0x7F);
  ZipAdd(hz,"from-mem.dat", buf4,1000,ZIP_MEMORY);

  // (10) add from a pipe
  CreatePipe(&hread,&hwrite,NULL,0);
  hthread = CreateThread(NULL,0,WriterFunc,(void*)hwrite,0,&tid);
  ZipAdd(hz,"from-pipe.dat",  hread,0,ZIP_HANDLE);
  // nb. WriterFunc closes its hwrite handle itself. That's so that the
  // call to ZipAdd can recognise the end-of-data and terminate.
  WaitForSingleObject(hthread,INFINITE);
  CloseHandle(hthread);
  CloseHandle(hread);


  CloseZip(hz);



  MessageBox(NULL,"Testing zipfile reading...","Ziptest",MB_OK);


  // (1) open from a filename
  hz = OpenZip("ziptest-fromstuff.zip",0,ZIP_FILENAME);
  ZIPENTRY ze; GetZipItem(hz,-1,&ze); int numitems=ze.index;
  for (int i=0; i<numitems; i++)
  { GetZipItem(hz,i,&ze);
    UnzipItem(hz,i, ze.name,0,ZIP_FILENAME);
  }
  CloseZip(hz);

  // (2) open from a file handle
  hf = CreateFile("ziptest-fromstuff.zip",GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
  hz = OpenZip(hf,0,ZIP_HANDLE);
  UnzipItem(hz,1,"unztest-fromhandle.dat",0,ZIP_FILENAME);
  CloseZip(hz);

  // (3) open from a pipe handle
  // !!! not yet implemented

  // (4) open from a memory block
  hf = CreateFile("ziptest-fromstuff.zip",GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
  DWORD size = GetFileSize(hf,0); char *buf5=new char[size];
  DWORD red; ReadFile(hf,buf5,size,&red,NULL);
  CloseHandle(hf);
  hz = OpenZip(buf5,size, ZIP_MEMORY);
  UnzipItem(hz,1,"unztest-frommemory.dat",0,ZIP_FILENAME);
  CloseZip(hz);
  delete[] buf5;



  hz = OpenZip("ziptest-fromstuff.zip",0,ZIP_FILENAME);

  // (5) find an item.
  int index; FindZipItem(hz,"from-mem.dat",false,&index,&ze);
  UnzipItem(hz,index,"unztest-found.dat",0,ZIP_FILENAME);

  // (6) to a filename
  UnzipItem(hz,1,"unztest-tofilename.dat",0,ZIP_FILENAME);

  // (7) to a file handle
  GetZipItem(hz,1,&ze);
  if ((ze.attr&FILE_ATTRIBUTE_DIRECTORY)==0)
  { hf = CreateFile("unztest-tofilehandle",GENERIC_WRITE,0,NULL,CREATE_ALWAYS,ze.attr,0);
    UnzipItem(hz,1,hf,0,ZIP_HANDLE);
    SetFileTime(hf,&ze.ctime,&ze.atime,&ze.mtime);
    CloseHandle(hf);
  }

  // (8) to a memory block
  GetZipItem(hz,1,&ze);
  char *buf6 = new char[ze.unc_size];
  UnzipItem(hz,1,buf6,ze.unc_size,ZIP_MEMORY);
  hf = CreateFile("unztest-tomem",GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,0);
  WriteFile(hf,buf6,ze.unc_size,&writ,NULL);
  CloseHandle(hf);
  delete[] buf6;

  // (9) to a pipe
  CreatePipe(&hread,&hwrite,NULL,0);
  hthread = CreateThread(NULL,0,ReaderFunc2,(void*)hread,0,&tid);
  UnzipItem(hz,1,hwrite,0,ZIP_HANDLE);
  CloseHandle(hwrite);
  WaitForSingleObject(hthread,INFINITE);
  CloseHandle(hthread);
  CloseHandle(hread);



  CloseZip(hz);


  MessageBox(NULL,"Finished tests.","Ziptest",MB_OK);


  return 0;
}










